{-# LANGUAGE FlexibleContexts #-}

module Analysis where
import Echelon
import Datastructures
import Control.Exception (assert)
import qualified Data.Sequence as Seq 
import Data.Foldable (toList)
import Debug.Trace
import Control.Monad
import qualified Data.Map as Map
import qualified Data.Set as Set
import Data.List
import Data.Maybe
import Control.Monad.State

-- COLOR DERIVATION
-- add a color to the network data structure x.
-- Packet d has reached in-channel p of component c. The index of c in x is i.
add_color :: Show p_t => Eq p_t => p_t -> Component p_t -> Int -> Int -> Network p_t -> Maybe (Network p_t)
add_color d c i p x = case (prim c) of
						Source injs s -> Just x
						_ -> trace ("adding color " ++ show d ++ " to in " ++ show p ++ " of " ++ ide c) $
			 			  		case Seq.elemIndexL d (Seq.index (colors c) p) of
									Just n  -> trace "Already added." Nothing
									Nothing -> Just (Seq.update i (c { colors = Seq.update p ((Seq.<|) d (Seq.index (colors c) p)) (colors c) }) x)


first4  (x, _, _,_) = x

-- derive all colors by folding over all components and propagating each injected packet at each source
derive_colors :: Show p_t => Eq p_t => Network p_t -> Network p_t
derive_colors = join derive_colors_rec

	where
		derive_colors_rec x x' = Seq.foldrWithIndex(\n c x' -> case prim c of
												Source injs s -> Prelude.foldr (\inj x' -> first4 (derive_colors_one inj n 0 x')) x' injs
												_ -> x')
									x'
									x
		derive_colors_one d i p x =
			let c = Seq.index x i in
				--trace ((ide c) ++ "  " ++ show d) $ 
				case add_color d c i p x of
					Just x -> foldr (goto_out c) (x, Nothing, p, d)
						(case prim c of
							Queue n	    -> [(0,d)] 
							Merge       -> [(0,d)] 
							Switch f s  -> [if f d then (0,d) else (1,d)]
							ASwitch fs  -> (if appASwitchFunction (fs!!0) d then [(0,d)] else []) ++ (if appASwitchFunction (fs!!1) d then [(1,d)] else [])
							Source inj s-> [(0,d)] 
							Fork		-> [(0,d),(1,d)]
							Join di		-> if p == di then [(0,d)] else []
							Sink dead	-> []
							Function f s-> [(0,f d)] 
							Peek		-> [(0,d)]
							Automaton ss token ts-> map fromJust $ filter (/= Nothing) $ map (execTransition d p) ts
						)
					Nothing -> (x, Nothing, p, d)
		goto_out c (p',d') (x,s,p,d) = case (Seq.index (outs c) p') of
										Just (Left i,p) -> derive_colors_one d' i p x
										Just (Right s,p) -> assert False (x, Just s, p, d') -- TODO hoeft niet, kan gewoon doorgaan
										_ -> (x, s, p, d')
		execTransition d p (s,s',e,f) = if e p d then f p d else Nothing


-- INVARIANTS

-- a channel is the index of a component + the outport
type Channel = (Int, Int)

-- Returns a Channel from an in-port of a component
-- Assumes the initiator of the in-port exists and that blanks have been filled.
in_to_Channel :: Network p_t -> Component p_t -> Int -> Channel
in_to_Channel x c p = case Seq.index (ins c) p of
						Just (Left i, p)    -> (i,p)
						Just (Right idx, p) -> case Seq.findIndexL (\c -> ide c == idx) x of
													Just i -> (i,p)

only_if :: Bool -> (a -> a) -> a -> a
only_if cond f a = if cond then f a else a

-- Adds a Channel to the channel map, without overwriting existing values
add_if_new :: Channel -> Map.Map Channel Int -> Map.Map Channel Int
add_if_new ch ch_map = only_if (not $ Map.member ch ch_map)
							   (Map.insert ch (Map.size ch_map))
							   ch_map

-- Creates a mapping from Channels to Integers
-- Assumes blanks in $x$ have been filled (see function fill_blanks)

create_channel_map_naive :: Network p_t -> Map.Map Channel Int
create_channel_map_naive x = Seq.foldrWithIndex add_transitions (Seq.foldrWithIndex add_ins_and_outs Map.empty x) x
	where
		add_ins_and_outs i c ch_map = let ch_map' = Seq.foldrWithIndex (\i -> add_if_new)
														ch_map $ Seq.mapWithIndex (\p i -> in_to_Channel x c p) $ ins c in
													  Seq.foldrWithIndex (\i -> add_if_new)
														ch_map $ Seq.mapWithIndex (\p o -> (i,p)) $ outs c
		add_transitions i c ch_map = case prim (Seq.index x i) of
										Automaton ss token ts -> foldr add_if_new ch_map $ map (\n -> (i,n)) [Seq.length (outs c)..Seq.length (outs c) + length ts]
										_ -> ch_map

-- Equates two channels in the channel mapping
-- Channels $ch1$ now points to $ch0$
equate_channels :: Channel -> Channel -> Map.Map Channel Int -> Map.Map Channel Int
equate_channels ch0 ch1 ch_map = let i0 = ch_map Map.! ch0;
									 i1 = ch_map Map.! ch1 in
									Map.map (\a -> if a == i1 then i0 else a) ch_map


-- Updates a channel map by equating the channels of forks and the data channels of joins
-- Assumes blanks in $x$ have been filled (see function fill_blanks)
equate_joins_forks :: Network p_t -> Map.Map Channel Int -> Map.Map Channel Int
equate_joins_forks x ch_map = Seq.foldrWithIndex set_join_channels ch_map x
	where set_join_channels i c ch_map  = case (prim c) of
											Join di -> -- the data in-channel
													   let ch_i = in_to_Channel x c di;
													    -- the out-channel
														ch_o  = (i,0) in
														-- Replace all occurences of $o$ with $i$
														equate_channels ch_i ch_o ch_map
											Fork    ->  -- the in-channel
													   let ch_i = in_to_Channel x c 0;
													    -- the out-channels
														ch_o0 = (i,0);
														ch_o1 = (i,1) in
														-- Replace all occurences of $o0$ and $o1$ with $i$
														equate_channels ch_i ch_o0
														$ equate_channels ch_i ch_o1 ch_map
											_ -> ch_map
														
create_channel_map x = equate_joins_forks x (create_channel_map_naive x)


get_trgt :: Network p_t -> Component p_t -> Int -> Maybe (Component p_t,Int)
get_trgt x c p = case Seq.index (outs c) p of
					Just (Left i,p')    -> Just (Seq.index x i,p')
					Just (Right idx,p') -> case Seq.findIndexL (\c -> ide c == idx) x of
											Just i  -> Just (Seq.index x i,p')
											Nothing -> Nothing
					Nothing -> Nothing
					

-- Create a matrix of invariants without equalities.
-- The fork induces two equalities (all three channels are equal) and the join one (the data-in is equal to the out)
-- These equalities are not added as invariants in the matrix, since we will map channels to integers in such a way that equal channels map to the same integer.
-- Assumes colors have been derived.
-- If the target of a component is Nothing or unknown, then we assume a fair sink.
-- A peek is basically a dead sink, since its trdy is always low.
foldInd :: (Int -> a -> b -> b) -> b -> [a] -> b
foldInd = foldIndN 0
	where
		foldIndN n f b [] = b
		foldIndN n f b (a:as) = foldIndN (n+1) f (f n a b) as




partition_ins :: Show p_t => Eq p_t => [(Int,p_t)] -> [AutomatonT p_t] -> [([(Int,p_t)],[Int])] -> [([(Int,p_t)],[Int])]
partition_ins [] 		  ts mp = let ts' = filter (\t -> not $ t `elem` (concat $ snd $ unzip mp)) [0..(length ts)-1] in
									if null ts' then mp else ([], ts'):mp
partition_ins ((p,d):ins) ts mp = 
		let ts'  = filter (\n -> isTriggered p d (ts!!n)) [0..(length ts)-1];
			cs   = [(p,d)];
			mp' = case find (sharedTransition ts') mp of
					Nothing -> (cs,ts') : mp
					Just (k,a) -> (cs ++ k,nub (ts'++a)) : (delete (k,a) mp) in
			partition_ins ins ts mp'
	where
		isTriggered p d (s,s',e,f) = e p d
		sharedTransition ts (k,a)  = ts `intersect` a /= [] 


partition_outs :: Eq p_t => Component p_t -> [(Int,p_t)] -> [AutomatonT p_t] -> [([(Int,p_t)],[Int])] -> [([(Int,p_t)],[Int])]
partition_outs c [] 		  ts mp = mp
partition_outs c ((p,d):outs) ts mp = 
		let ts' = filter (\n -> canProduce p d (ts!!n)) [0..(length ts)-1];
			mp' = case find (sharedTransition ts') mp of
					Nothing -> ([(p,d)],ts') : mp
					Just (k,a) -> ((p,d):k,nub (ts'++a)) : (delete (k,a) mp) in
			partition_outs c outs ts mp'
	where
		canProduce p d (s,s',e,f) = or (map (\(p',color') -> (e p' color') && f p' color' == Just (p,d)) ins_with_colors)
		ins_with_colors = concat $ map (\p' -> map (\color -> (p',color)) (toList $ Seq.index (colors c) p')) [0..Seq.length (ins c)-1];
		sharedTransition ts (k,a) = ts `intersect` a /= [] 



partition_outs2 :: Eq p_t => Network p_t -> Component p_t -> [(Int,p_t)] -> [AutomatonT p_t] -> [([(Int,p_t)],[(Int,p_t)])] -> [([(Int,p_t)],[(Int,p_t)])]
partition_outs2 x c [] 		  ts mp = let outs = filter (\(p,color) -> not $ (p,color) `elem` (concat $ fst $ unzip mp)) outs_with_colors;
										  ins  = filter (\(p,color) -> not $ (p,color) `elem` (concat $ snd $ unzip mp)) ins_with_colors;
										  mp' = if null outs then mp else (outs, []):mp;
										  mp'' = if null ins then mp else ([], ins):mp' in
											filter (\(outs,ins) -> case find inProducingNothing ins of
												Just i -> False
												Nothing -> True) mp''
	where
		inProducingNothing (p,d) = or (map (\(s,s',e,f) -> e p d && f p d == Nothing) ts)
		ins_with_colors = concat $ map (\p' -> map (\color -> (p',color)) (toList $ Seq.index (colors c) p')) [0..Seq.length (ins c)-1];
		outs_with_colors = concat $ map (\p' -> map (\color -> (p',color)) (toList $ trgt_colors c p')) [0..Seq.length (outs c)-1]
		trgt c p = case Seq.index (outs c) p of {Just (Left i,p') -> (i,p')}
		trgt_colors c p = let (i',p') = trgt c p in Seq.index (colors (Seq.index x i')) p'
partition_outs2 x c ((p,d):outs) ts mp =  
		let ins = filter (\(p',color') -> canProduce p d (p',color')) ins_with_colors;
			mp' = case find (sharedIn ins) mp of
					Nothing -> ([(p,d)],ins) : mp
					Just (k,a) -> ((p,d):k,nub (ins++a)) : (delete (k,a) mp) in
			partition_outs2 x c outs ts mp'
	where
		canProduce p d (p',color') = or (map (\(s,s',e,f) -> (e p' color') && f p' color' == Just (p,d)) ts)
		ins_with_colors = concat $ map (\p' -> map (\color -> (p',color)) (toList $ Seq.index (colors c) p')) [0..Seq.length (ins c)-1];
		sharedIn ins' (k,a) = ins' `intersect` a /= [] 


create_invariant_matrix_wo_equalities :: Show p_t => Ord p_t => Network p_t -> [(Map.Map (Channel, p_t) Int, Map.Map (Either (Int,p_t) (Int,Int)) Double)]
create_invariant_matrix_wo_equalities x = Seq.foldrWithIndex create_invariant_per_component [] x
	where create_invariant_per_component i c matrix = matrix ++
		case prim c of
			Queue s  	  -> map queue_invariant (toList $ Seq.index (colors c) 0)
			Join di	 	  -> join_invariant_ti (if di==0 then 1 else 0)
			Switch f fs	  -> map (switch_invariant f) (toList $ Seq.index (colors c) 0)
			ASwitch fs	  -> map (aswitch_invariant fs) (toList $ Seq.index (colors c) 0)
			Merge    	  -> case get_trgt x c 0 of
								Just (trgt,p') -> map merge_invariant (toList $ Seq.index (colors trgt) p')
								Nothing		   -> []
			Function f fs -> case get_trgt x c 0 of
								Just (trgt,p') -> map (function_invariant f) (toList $ Seq.index (colors trgt) p')
								Nothing		   -> []
			Sink True	  -> map dead_sink_invariant (toList $ Seq.index (colors c) 0)
			Peek		  -> map dead_sink_invariant (toList $ Seq.index (colors c) 0)
			Automaton ss token ts -> let offset = (Seq.length $ outs c) in
									--	[automaton_invariant ts offset token] ++
										(map (automaton_ts_invariant ts offset token) [0..(length $ getStates ts)-1]) ++
										(map (automaton_i_invariant offset token) (partition_ins    (toList $ ins_with_colors)  ts [])) ++
										(map (automaton_o_invariant offset token) (partition_outs c (toList $ outs_with_colors) ts [])) ++
										(map (automaton_oi_invariant offset token) (partition_outs2 x c (toList $ outs_with_colors) ts []))
			_ -> []
			where {
				  automaton_invariant ts offset token = 
												let map0 = foldr (\s -> Map.insert (Right (i,s+1)) (-1)) Map.empty [0..(length $ getStates ts)-1];
													map1 = Map.insert (Right (0, 0)) 1 map0 in
													(Map.empty, map1);
				  automaton_oi_invariant offset token (outs,ins) =
													let map0 = foldr (\(p',d') -> Map.insert (in_to_Channel x c p', d') 1) Map.empty ins;
														map1 = foldr (\(p,d)   -> Map.insert ((i,p), d) (-1)) (assert (map0 /= Map.empty) map0) outs in
													(map1,Map.empty);
				  automaton_i_invariant offset token (ins,ts) =
													-- For each partition of in-channels and transitions:
													-- + the sum of all ins
													let map0 = foldr (\(p,color) -> Map.insert (in_to_Channel x c p, color) 1) Map.empty ins;
													-- - the sum of all transitions triggered by these ins
														map1 = foldr (\n -> Map.insert ((i,n+offset), token) (-1)) map0 ts in
													 (map1,Map.empty);
				  automaton_o_invariant offset token (outs,ts) =
													-- For each partition of out-channels and transitions:
													-- + the sum of all outs
													let map0 = foldr (\(p,color) -> Map.insert ((i, p), color) 1) Map.empty outs;
													-- - the sum of all transitions that may produce the packet at these outs
														map1 = foldr (\n -> Map.insert ((i,n+offset), token) (-1)) map0 ts in
													 (map1,Map.empty);
				  ins_with_colors  = concat $ map (\p' -> map (\color -> (p',color)) (toList $ Seq.index (colors c) p')) [0..Seq.length (ins c)-1];
				  outs_with_colors = concat $ map (\p' -> map (\color -> (p',color)) (toList $ trgt_colors c p')) [0..Seq.length (outs c)-1];
				  trgt c p = case Seq.index (outs c) p of {Just (Left i,p') -> (i,p')};
				  trgt_colors c p = let (i',p') = trgt c p in Seq.index (colors (Seq.index x i')) p';
				  automaton_ts_invariant ts offset token s =
												-- + the sum of all ins of state s:
											    let map0 = foldInd (\n (s',s'',e,f) -> only_if (s'' == s && s' /= s) (Map.insert ((i,n+offset),token) 1)) Map.empty ts;
											    -- - the sum of all outs of state s:
											    	map1 = foldInd (\n (s',s'',e,f) -> only_if (s' == s && s'' /= s) (Map.insert ((i,n+offset),token) (-1))) map0 ts;
												-- - the state
													map2 = Map.singleton (Right (i, s+1)) (-1);
												-- + 1 if the state is the initial state
													map3 = if s == 0 then Map.singleton (Right (0, 0)) 1 else Map.empty in
												 (map1, Map.union map2 map3);
				  dead_sink_invariant color	 = -- +c.color
				  							   (Map.singleton (in_to_Channel x c 0, color) 1, Map.empty);
				  function_invariant f color = (-- for any color $color'$ of the in-channel that yields $color$ on the output: + f_in.color'
											    let map = Seq.foldrWithIndex (\n color' -> only_if (f color' == color) (Map.insert (in_to_Channel x c 0, color') 1))
																		  Map.empty (Seq.index (colors c) 0) in
											      -- - f_out.color
												  Map.insert ((i,0),color) (-1) $ assert (map /= Map.empty) map,
										        Map.empty);
				  merge_invariant color		 = (-- for any in-channel that contains the given color: + mrg_in.color
											    let map = Seq.foldrWithIndex (\p colors -> only_if (Seq.elemIndexL color colors /= Nothing)
																							   (Map.insert (in_to_Channel x c p, color) 1))
  																	   	  Map.empty (colors c) in
											    -- - mrg_out.color
												  Map.insert ((i,0),color) (-1) $ assert (map /= Map.empty) map,
										        Map.empty);
				  switch_invariant f color	 = (Map.insert (in_to_Channel x c 0, color) 1 							-- + sw_in.color
												$ Map.insert ((i,if f color then 0 else 1), color) (-1) Map.empty,	-- - sw_out.color
											    Map.empty);
				  aswitch_invariant fs color = (Map.insert (in_to_Channel x c 0, color) 1 											-- + sw_in.color
												$ Map.fromList (
													(if appASwitchFunction (fs!!0) color then [(((i,0),color), (-1))] else []) ++	-- - sw_out0.color
													(if appASwitchFunction (fs!!1) color then [(((i,1),color), (-1))] else [])),	-- - sw_out1.color
												Map.empty);
				  join_invariant_ti ti		 = [assert (Seq.length (Seq.index (colors c) 0) > 0 && Seq.length (Seq.index (colors c) 1) > 0) (
												(let di  = if ti == 0 then 1 else 0;
												   -- sum over colors of ti ( + join_ti.color )
												     map = Seq.foldrWithIndex (\n color -> Map.insert (in_to_Channel x c ti, color) 1) Map.empty (Seq.index (colors c) ti) in
												   -- sum over colors of di ( - join_di.color )
												  Seq.foldrWithIndex (\n color -> Map.insert (in_to_Channel x c di, color) (-1)) map (Seq.index (colors c) di),
											  Map.empty))];
    			  queue_invariant color		 = (Map.insert (in_to_Channel x c 0, color) 1		-- + q_in.color
				 								$ Map.insert ((i,0), color) (-1) Map.empty,		-- - q_out.color
										        Map.insert (Left (i,color)) (-1) Map.empty)		-- - q.color
			}

-- Creates a matrix of invariants.
-- First, the matrix without equalities is generated.
-- Then the channels are mapped to integers using the channel map.
-- Equal channels (in case of the fork and join) map to equal integers.
-- Assumes colors have been derived and blanks have been filled.
create_invariant_matrix :: Show p_t => Ord p_t => Network p_t ->  [(Map.Map (Int, p_t) Int, Map.Map (Either (Int,p_t) (Int,Int)) Double)]
create_invariant_matrix x = let ch_map = create_channel_map x;
						  		matrix = create_invariant_matrix_wo_equalities x in
							  map (\mat_row -> (Map.mapKeys (\ch_pa -> (case Map.lookup (fst ch_pa) ch_map of
																				Nothing -> trace ((show $ fst ch_pa) ++ show (ide (Seq.index x (fst $ fst ch_pa)))) assert False 0
																				Just i -> i
																		, snd ch_pa)) $ fst mat_row, snd mat_row)) matrix



map_lefts :: Ord a => Map.Map (Either a b) v -> Map.Map a v
map_lefts s = Map.mapKeys (\ab -> case ab of Left a -> a)
			$ Map.filterWithKey (\ab v -> case ab of
									Left a  -> True
									Right b -> False) s
map_rights :: Ord b => Map.Map (Either a b) v -> Map.Map b v
map_rights s = Map.mapKeys (\ab -> case ab of Right b -> b)
			$ Map.filterWithKey (\ab v -> case ab of
									Left a  -> False
									Right b -> True) s

-- simplify the matrix by introducing variables of type q0 representing all packets in q0.
-- Whenever q0.d0 + q0.d1 + ... is present in an equation replace it with q0 (with d0, d1, ... the complete set of derived colors of q0)
-- The same for negative values.
-- assumes values in the matrix are always 1 or -1
simplify_using_derived_colors :: Ord p_t => Network p_t -> Map.Map (Either (Int,p_t) (Int,Int)) Int -> (Map.Map (Int,p_t) Int, Map.Map (Either Int (Int,Int)) Int)
simplify_using_derived_colors x row = Seq.foldrWithIndex simplify_using_derived_colors_one (map_lefts row, Map.mapKeys Right $ map_rights row) x
	where simplify_using_derived_colors_one i q row =
		case prim q of
			Queue s -> let row' = if Seq.foldrWithIndex (\n color b -> b && Map.lookup (i,color) (fst row) == Just 1) (Seq.length (Seq.index (colors q) 0) > 0) (Seq.index (colors q) 0) then
							(Seq.foldrWithIndex (\n color -> Map.delete (i,color)) (fst row) (Seq.index (colors q) 0),
							 Map.insert (Left i) 1 (snd row))
				 	    else
							row in
						if Seq.foldrWithIndex (\n color b -> b && Map.lookup (i,color) (fst row') == Just (-1)) (Seq.length (Seq.index (colors q) 0) > 0) (Seq.index (colors q) 0) then
							(Seq.foldrWithIndex (\n color -> Map.delete (i,color)) (fst row') (Seq.index (colors q) 0),
							 Map.insert (Left i) (-1) (snd row'))
				 	    else
							row'
			_		-> row

-- Create invariants:
-- First, create a matrix containing lambda variables and the left side and q0.d variables on the right side.
-- Using echelonFilter, sweep this matrix, eliminating all lambda's.
-- Then simplify the matrix using the derived colors.
create_invariants :: Show p_t => Ord p_t => Network p_t -> [(Map.Map (Int,p_t) Int, Map.Map (Either Int (Int,Int)) Int)]
create_invariants x = let ret1 = map (Map.map round) (echelonFilter $ create_invariant_matrix x) in
						map (simplify_using_derived_colors x) ret1



