{-# LANGUAGE DeriveDataTypeable #-}

module SMT where
import Datastructures
import Analysis

import System.Environment
import Control.Exception (assert)
import qualified Data.Sequence as Seq
import Data.Foldable (toList)
import Debug.Trace
import Control.Monad
import qualified Data.Map as Map
import Data.Data
import Data.Maybe
import Data.Text (split)
import Data.Char (isDigit)
import Data.String.Conversions
import Data.List
import Echelon
import System.IO
import qualified Data.ByteString.Char8 as B
import qualified Data.Set as Set
import Text.Parsec
import Text.Parsec.String
import Data.Either

-- EXPORTING TO SMT
export_value_to_smt :: Int -> String
export_value_to_smt v = if v > 0 then show v else "( - " ++ show (-v) ++ ")" -- if v > 0 then "" else if v < 0 then "- " else traceShow v $ assert False (if v >= 0 then "" else "- ")

export_q_to_smt :: Network p_t -> (Int, Int) -> String
export_q_to_smt x (q, v) = "( * " ++ export_value_to_smt v ++ " " ++ ide (Seq.index x q) ++ ")"

export_state_to_smt :: Network p_t -> ((Int, Int), Int) -> String
export_state_to_smt x ((a, 0), v) = assert (a==0) "( * " ++ export_value_to_smt v ++ " 1)"
export_state_to_smt x ((a, n), v) = case prim $ Seq.index x a of 
										Automaton ss token ts -> 
											"( * " ++ export_value_to_smt v ++ " " ++ ide (Seq.index x a) ++ (ss!!(n-1)) ++ ")"

export_q_data_to_smt :: Network p_t -> (p_t -> String) -> ((Int, p_t), Int) -> String
export_q_data_to_smt x show_p_t ((q, d), v) = "( * " ++ export_value_to_smt v ++ " " ++ ide (Seq.index x q) ++ "." ++ show_p_t d ++ ")"

export_q_or_state_to_smt :: Network p_t -> (Either Int (Int, Int),Int) -> String
export_q_or_state_to_smt x ((Left q),v) = export_q_to_smt x (q,v)
export_q_or_state_to_smt x ((Right (a,n)),v) = export_state_to_smt x ((a,n),v)

export_row_to_smt :: Network p_t -> (p_t -> String) -> (Map.Map (Int, p_t) Int, Map.Map (Either Int (Int,Int)) Int) -> String
export_row_to_smt x show_p_t row =
	if not (Map.null (fst row) && Map.null (snd row)) then
						"(assert (= 0 (+ "
						  ++ intercalate " " (map (export_q_data_to_smt x show_p_t) (Map.toList (fst row))
											++
										     (map (export_q_or_state_to_smt x) $ Map.toList (snd row)))
						  ++ ")))"
	else ""

export_q_data_var_to_smt :: Network p_t -> (p_t -> String) -> (Int, p_t) -> String
export_q_data_var_to_smt x show_p_t (i,d) = let q = Seq.index x i in
							case prim q of Queue s -> "(declare-fun " ++ ide q ++ "." ++ show_p_t d ++ " () Int) (assert (<= 0 " ++ ide q ++ "." ++ show_p_t d ++ "))"

export_q_var_to_smt :: Network p_t -> Int -> String
export_q_var_to_smt x i = let q = Seq.index x i in
							case prim q of
								Queue s -> "(declare-fun " ++ ide q ++ " () Int) (assert (<= 0 " ++ ide q ++ "))(assert (<= " ++ ide q ++ " " ++ show s ++ "))"
								Merge   -> "(declare-fun " ++ ide q ++ " () Bool)"

export_state_var_to_smt :: Network p_t -> (Int,Int) -> String
export_state_var_to_smt x (0,0) = ""
export_state_var_to_smt x (i,n) = let a = Seq.index x i in
									case prim a of
									Automaton ss token f ->
											let idx = ide a ++ (ss!!(n-1)) in assert (n/=0) $
											"(declare-fun " ++ idx ++ " () Int) (assert (<= 0 " ++ idx ++ "))(assert (<= " ++ idx ++ " 1))"			

bi_to_name :: Network p_t -> (p_t -> String) -> (Literal p_t) -> String -- TODO queues, srcs no trailing ".0"
bi_to_name x show_p_t (Block i p Nothing) = let c = Seq.index x i in
												"block_" ++ ide c ++ "." ++ show p
bi_to_name x show_p_t (Idle i p Nothing) = let c = Seq.index x i in
												(case prim c of Queue s -> "idle_"; _ -> "dead_") ++ ide c ++ "." ++ show p
bi_to_name x show_p_t (Block i p (Just d)) = let c = Seq.index x i in
												"block_" ++ ide c ++ "." ++ show p ++ "." ++ show_p_t d
bi_to_name x show_p_t (Idle i p (Just d)) = let c = Seq.index x i in
												(case prim c of Queue s -> "idle_"; _ -> "dead_") ++ ide c ++ "." ++ show p ++ "." ++ show_p_t d

export_bi_var_to_smt :: Network p_t -> (p_t -> String) -> (Literal p_t) -> String
export_bi_var_to_smt x show_p_t lit = "(declare-fun " ++ bi_to_name x show_p_t lit ++ " () Bool)"

export_color_info_to_smt :: Network p_t -> (p_t -> String) -> Int -> String
export_color_info_to_smt x show_p_t i = let q = Seq.index x i;
											ds = toList (Seq.index (colors q) 0) in
											if length ds == 1 then
												"(assert (= " ++ ide q ++ " " ++ ide q ++ "." ++ show_p_t (ds!!0) ++ "))"
											else
												"(assert (= " ++ ide q ++  " (+ "
												++ (intercalate " " (map (\d -> ide q ++ "." ++ show_p_t d) $ ds))
												++ ")))"


-- Export the invariants to SMT
-- First, create them. Then assemble all variables in two sets (one for variables q0.d, others for variables q0).
-- Examine the first set to find out which queues are color specific, i.e., have invariants over them with q0.d variables.
-- For those queues, the color derivation is exported to SMT as well.
export_invariants_to_smt :: (Show p_t,Ord p_t) => Network p_t -> (p_t -> String) -> (String,Set.Set Int, (Set.Set (Int, p_t), Set.Set Int))
export_invariants_to_smt x show_p_t =
	let inv   = create_invariants x;
	    vars  = get_variables_in_invariants inv;
		states= get_states_in_invariants inv;
		qs    = get_color_specific_queues (fst vars);
		vars' = add_queues_and_colors vars qs;
		s     = intercalate "\n" (map (export_q_var_to_smt x) $ Set.toList (snd vars')) ++
				intercalate "\n" (map (export_q_data_var_to_smt x show_p_t) $ Set.toList (fst vars')) ++ "\n" ++
				intercalate "\n" (map (export_state_var_to_smt x) (Set.toList states)) ++
				"\n\n" ++ 
				intercalate "\n" (map (export_color_info_to_smt x show_p_t) $ Set.toList qs) ++
				"\n\n" ++ 
				intercalate "\n" (map (export_row_to_smt x show_p_t) inv) in
		(s,qs, vars')
	where {
		get_variables_in_invariants = foldr (\row vars -> ((Set.union (Map.keysSet (fst row)) (fst vars)),
												   		   (Set.union (Map.keysSet (map_lefts $ snd row)) (snd vars))))
											(Set.empty,Set.empty);
		get_states_in_invariants = foldr (\row vars -> (Set.union (Map.keysSet (map_rights $ snd row)) vars))
											Set.empty;
		get_color_specific_queues vars = foldr (\q_d -> Set.insert $ fst q_d) Set.empty $ Set.toList vars;
		add_queues_and_colors = Set.foldr (\q vars -> (Set.union (Set.fromList (map (\color -> (q,color)) (toList (Seq.index (colors (Seq.index x q)) 0)))) (fst vars)
													,
												   Set.insert q (snd vars)))
	}





-- Export formulas to SMT
export_literal_to_SMT :: (Show p_t,Ord p_t) => Network p_t -> (p_t -> String) -> Literal p_t -> String
export_literal_to_SMT x show_p_t (Is_Full q) = case prim (Seq.index x q) of
												Queue s -> "(= " ++ show s ++ " " ++ (export_q_to_smt x (q,1)) ++ ")"
export_literal_to_SMT x show_p_t (Is_Gt_One q d) = "(>= " ++ (case d of
																	Nothing -> export_q_to_smt x (q,1) 
																	Just d' -> export_q_data_to_smt x show_p_t ((q,d'),1))
														++ " 1)"
export_literal_to_SMT x show_p_t (Is_Zero q d) = "(= " ++ (case d of
																	Nothing -> export_q_to_smt x (q,1) 
																	Just d' -> export_q_data_to_smt x show_p_t ((q,d'),1))
														++ " 0)"
export_literal_to_SMT x show_p_t (Sum_Compare [(q,d)] cmp n) = "(" ++ cmp ++ " " ++ (case d of
																	Nothing -> export_q_to_smt x (q,1) 
																	Just d' -> export_q_data_to_smt x show_p_t ((q,d'),1))
														++ " " ++ show n ++ ")"
export_literal_to_SMT x show_p_t (Sum_Compare qs cmp n) = "(" ++ cmp ++ " (+ " ++ (intercalate " " (map (\(q,d) -> (case d of
																	Nothing -> export_q_to_smt x (q,1) 
																	Just d' -> export_q_data_to_smt x show_p_t ((q,d'),1))) qs)) ++ ")"
														++ " " ++ show n ++ ")"
export_literal_to_SMT x show_p_t (InState a s) = case prim (Seq.index x a) of
													Automaton ss token ts -> "(= 1 " ++ (export_state_to_smt x ((a,s+1),1)) ++ ")"
export_literal_to_SMT x show_p_t (NotInState a s) = case prim (Seq.index x a) of
														Automaton ss token ts -> "(= 0 " ++ (export_state_to_smt x ((a,s+1),1)) ++ ")"
export_literal_to_SMT x show_p_t (Select i b)  = "(= " ++ (if b then "true" else "false") ++ " " ++ ide (Seq.index x i) ++ ")"
export_literal_to_SMT x show_p_t (Block i p d) = bi_to_name x show_p_t (Block i p d)
export_literal_to_SMT x show_p_t (Idle i p d)  = bi_to_name x show_p_t (Idle i p d)



export_formula_to_SMT :: (Show p_t,Ord p_t) => Network p_t -> (Set.Set Int) -> (Set.Set (Int, p_t), Set.Set Int) -> (p_t -> String) -> Maybe (Literal p_t) -> Formula p_t -> (String,Set.Set Int, (Set.Set (Int, p_t), Set.Set Int))
export_formula_to_SMT x qs vars show_p_t bi f =
		let (s,vars',qs') = export_formula_to_SMT_rec $ simplify_singletons f;
			vars'' = Set.difference vars' qs';
			ret_vars = (Set.union (fst vars) (colors_from qs'), Set.union (snd vars) $ Set.union vars' qs');
			ret_qs   = (Set.union qs qs');
			ret_s = (Set.fold (++) "" (Set.map (\v -> export_q_var_to_smt x v ++ "\n") vars''))
				++
				(Set.fold (++) "" (Set.map (\v -> export_q_var_to_smt x v ++ "\n") (Set.difference qs' (snd vars))))
				++
				(Set.fold (++) "" (Set.map (\v -> export_q_data_var_to_smt x show_p_t v ++ "\n") (Set.difference (colors_from qs') (fst vars))))
				++
				(Set.fold (++) "" (Set.map (\q -> export_color_info_to_smt x show_p_t q ++ "\n") qs'))
				++ "\n" ++
				(case bi of
					Nothing -> "(assert " ++ s ++ ")"
					Just bi -> "(assert (= " ++ bi_to_name x show_p_t bi ++ " " ++ s ++ "))")
				++ "\n"
				in
			(ret_s,ret_qs,ret_vars)
	where
		export_formula_to_SMT_rec T = ("true",Set.empty,Set.empty)
		export_formula_to_SMT_rec F = ("false",Set.empty, Set.empty)
		export_formula_to_SMT_rec (Lit l) = (export_literal_to_SMT x show_p_t l, fst (new_vars l), snd (new_vars l))
		export_formula_to_SMT_rec (AND fs) = let (s',vars',qs') = unzip3 $ toList $ Set.map export_formula_to_SMT_rec fs in
												("(and " ++ intercalate " " s' ++ ")", foldl Set.union Set.empty vars', foldl Set.union Set.empty qs')
		export_formula_to_SMT_rec (OR fs)  = let (s',vars',qs') = unzip3 $ toList $ Set.map export_formula_to_SMT_rec fs in
												("(or "  ++ intercalate " " s' ++ ")", foldl Set.union Set.empty vars', foldl Set.union Set.empty qs')
		-- Returns a tuple (qs0,qs1) with qs0 new queue variables and qs1 new color-specific queue variables
		new_vars l = case l of
					Select i b -> (if i `Set.notMember` (snd vars) then Set.singleton i else Set.empty, Set.empty)
					Is_Full q -> (if q `Set.notMember` (snd vars) then Set.singleton q else Set.empty, Set.empty)
					Is_Gt_One q d -> (if q `Set.notMember` (snd vars) then Set.singleton q else Set.empty, Set.empty)
					Is_Zero q Nothing  -> (if q `Set.notMember` (snd vars) then Set.singleton q else Set.empty, Set.empty)
					Is_Zero q (Just d) -> if (q,d) `Set.notMember` (fst vars) then (Set.empty, Set.singleton q) else (Set.empty, Set.empty)
					Sum_Compare qs cmp v -> foldr (\(q,d) (l,r) -> case d of
																	Nothing -> (if q `Set.notMember` (snd vars) then Set.insert q l else l,r)
																	Just d  -> if (q,d) `Set.notMember` (fst vars) then (l, Set.insert q r) else (l,r))
											 (Set.empty, Set.empty) qs
					_ -> (Set.empty, Set.empty)
		colors_from qs' = Set.fold Set.union Set.empty (Set.map (\q -> Set.fromList (map (\color -> (q,color)) (toList $ Seq.index (colors (Seq.index x q)) 0))) qs')
		
		

-- Parsing SMT output into a model
pSMTUnsat :: Parser (Maybe SMTModel)
pSMTUnsat = do
			string "unsat"
			return Nothing

pSMTQ_Value :: Parser (String,Int)
pSMTQ_Value = do
			spaces
			string "(define-fun"
			spaces
			q_name <- manyTill anyChar (Text.Parsec.try (char ' '))
			spaces
			string "() "
			(string "Int" <|> string "Bool")
			spaces
			value <- intP
			string ")\n"
			return (q_name,value)
	where
		intP = (char '0' >> return 0) <|>
			   (string "false" >> return 0) <|>
			   (string "true" >> return 1) <|>
					(do
						firstDigit <- satisfy (\ ch -> '1' <= ch && ch <= '9')
						restDigits <- many digit
						return $ read $ firstDigit : restDigits
					) 

pSMTModel :: Parser (Maybe SMTModel)
pSMTModel = do
			spaces
			string "(model"
			ret <- manyTill pSMTQ_Value (char ')')
			return $ Just $ Map.fromList ret

pSMTSat :: Parser (Maybe SMTModel)
pSMTSat = do
			string "sat\n"
			pSMTModel

pSMTOutput :: Parser (Maybe SMTModel)
pSMTOutput = pSMTUnsat <|> pSMTSat

parseSMTOutput :: String -> Either ParseError (Maybe SMTModel)
parseSMTOutput s = parse pSMTOutput "(unknown)" s


showModel :: SMTModel -> String
showModel = Map.foldlWithKey (\s q v -> (if v > 0 && not (isPrefixOf "idle_" q || isPrefixOf "block_" q || isPrefixOf "dead_" q) then q ++ " := " ++ show v ++ "\n" else "") ++ s) "" 

